(in-package :contextl)

(defmacro deflayer (name &optional superlayers &body options)
  `(defclass ,name ,superlayers ,@(if options options '(()))))

(defgeneric layer-name (layer)
  (:method ((layer symbol)) layer)
  (:method ((layer class)) (class-name layer)))

(defgeneric find-layer (layer &optional errorp)
  (:method ((layer symbol) &optional (errorp t))
   (or (find-class layer nil)
       (when errorp
         (error "There is no layer named ~S." layer))))
  (:method ((layer class) &optional (errorp t))
   (declare (ignore errorp))
   layer))

(defun layer-prototype (layer)
  (class-prototype (find-layer layer)))

(defclass root-specializer () ())
(finalize-inheritance (find-class 'root-specializer))

(defstruct layer-context
  (prototype (class-prototype (find-class 'root-specializer))
             :type standard-object
             :read-only t)
  (specializer (find-class 'root-specializer)
               :type standard-class
               :read-only t)
  (children () :type list))

#-allegro
(declaim (type layer-context *active-context*))
#+allegro
(eval-when (:load-toplevel :execute)
  (proclaim '(type layer-context *active-context*)))

(defparameter *active-context*
  (make-layer-context))

(defun layer-active-p (layer &optional (context *active-context*))
  (subtypep (layer-context-specializer context)
            (find-layer layer)))

(defun maybe-create-layer-context (layer)
  (let ((layer-class (find-layer layer))
        (active-context-specializer (layer-context-specializer *active-context*)))
    (if (subtypep active-context-specializer layer-class)
        *active-context*
      (let ((new-specializer
             (make-instance 'standard-class
                            :direct-superclasses
                            (list layer-class active-context-specializer))))
        (finalize-inheritance new-specializer)
        (make-layer-context
         :prototype (class-prototype new-specializer)
         :specializer new-specializer)))))

(declaim (inline ensure-layer-context))

(defun ensure-layer-context (layer)
  (declare (optimize (speed 3) (debug 0) (safety 0)
                     (compilation-speed 0)))
  (or (getf (layer-context-children *active-context*) layer)
      (setf (getf (layer-context-children *active-context*) layer)
            (maybe-create-layer-context layer))))

(defun activate-layer (layer-name)
  (setf *active-context*
        (locally
          (declare (optimize (speed 3) (debug 0) (safety 0)
                             (compilation-speed 0)))
          (ensure-layer-context layer-name))))

(defmacro with-active-layer (layer-name &body body)
  (assert (symbolp layer-name))
  (if (eq layer-name 't)
      `(progn ,@body)
    `(let ((*active-context*
            (locally
              (declare (optimize (speed 3) (debug 0) (safety 0)
                                 (compilation-speed 0)))
              (ensure-layer-context ',layer-name))))
       ,@body)))

(defmacro with-active-layers ((&rest layer-names) &body body)
  (if layer-names
      `(with-active-layer ,(car layer-names)
         (with-active-layers ,(cdr layer-names)
           ,@body))
    `(progn ,@body)))

(defun funcall-in-layer (layer function &rest args)
  (declare (dynamic-extent args))
  (if (eq layer 't)
      (apply function args)
    (let ((*active-context*
           (locally
             (declare (optimize (speed 3) (debug 0) (safety 0)
                                (compilation-speed 0)))
             (ensure-layer-context layer))))
      (apply function args))))

(defun funcall-in-layers (layers function &rest args)
  (declare (dynamic-extent args))
  (if layers
      (funcall-in-layer
       (car layers)
       (lambda ()
         (apply #'funcall-in-layers (cdr layers) function args)))
    (apply function args)))

(defun apply-in-layer (layer function &rest args)
  (declare (dynamic-extent args))
  (if (eq layer 't)
      (apply #'apply function args)
    (let ((*active-context*
           (locally
             (declare (optimize (speed 3) (debug 0) (safety 0)
                                (compilation-speed 0)))
             (ensure-layer-context layer))))
      (apply #'apply function args))))

(defun apply-in-layers (layers function &rest args)
  (declare (dynamic-extent args))
  (if layers
      (funcall-in-layer
       (car layers)
       (lambda ()
         (apply #'apply-in-layers (cdr layers) function args)))
    (apply #'apply function args)))
